package org.netkernel.http.rest;

import java.io.ByteArrayInputStream;
import java.lang.reflect.Field;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.http.HttpMethod;
import org.netkernel.container.IKernel;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.MetaRepresentationImpl;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.LiteralParameterException;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.FastXPath;
import org.netkernel.layer0.util.NullOutputStream;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class RESTOverlayConfig implements IConfig
{
	private RESTOverlay mRESTOverlay;
	private ISpace mDelegateSpace;
	private ISpaceMeta mDelegateSpaceMeta;
	private String mBasePath;
	private Map<String, LogicalEndpointData> mLogicalEndpoints=new LinkedHashMap<String, LogicalEndpointData>();
	private boolean mStrict;
	private String mAuto404;
	private String mAuto406="";
	private String mAuto404EndpointIdentifier;
	private String mExceptionHandler="";
	private int mGlobalMaxContentLength=-1;
	private ISpaceElements mSpaceElements;
	private Map<String, String> mPassThroughEndpoints=new HashMap<String, String>();
	
	private static final String VERB_TRANSLATE_VERSION1="V1";

	public RESTOverlayConfig(ISpace aDelegateSpace, ISpaceMeta aDelegateSpaceMeta, INKFRequestContext aContext, RESTOverlay aRESTOverlay) throws Exception
	{	mDelegateSpace=aDelegateSpace;
		mDelegateSpaceMeta=aDelegateSpaceMeta;
		mRESTOverlay=aRESTOverlay;
		Document cfg=aContext.source("param:config", Document.class);
		mBasePath=XMLUtils.getText(FastXPath.getSingleNode(cfg.getDocumentElement(), "basepath"));
		mStrict=FastXPath.getSingleNode(cfg.getDocumentElement(), "strict")!=null;
		Node n404=FastXPath.getSingleNode(cfg.getDocumentElement(), "auto404");
		if(n404!=null)
		{	mAuto404=XMLUtils.getText(n404);				
		}
		if(mAuto404==null && n404!=null)
		{	mAuto404="";			
		}
		Node n406=FastXPath.getSingleNode(cfg.getDocumentElement(), "auto406");
		if(n406!=null)
		{	mAuto406=XMLUtils.getText(n406);
		}
		Node nExceptionHandler=FastXPath.getSingleNode(cfg.getDocumentElement(), "globalExceptionTarget");
		if(nExceptionHandler!=null)
		{	mExceptionHandler=XMLUtils.getText(nExceptionHandler);
		}
		Node nMaxContentSize=FastXPath.getSingleNode(cfg.getDocumentElement(), "maxContentLength");
		if(nMaxContentSize!=null)
		{	mGlobalMaxContentLength=Integer.parseInt(XMLUtils.getText(nMaxContentSize));				
		}
		aRESTOverlay.clearInitialisationProblems();
		initialize(aContext, aRESTOverlay);
	}

	public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
	{	return mSpaceElements;
	}

	public IMetaRepresentation getElementMeta(String aElementId)
	{	LogicalEndpointData led=mLogicalEndpoints.get(aElementId);
		return led.mEndpointMeta;
	}

	public String getResolvedElement(INKFResolutionContext aContext) throws Exception
	{	String result=null;
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		for (String key: mLogicalEndpoints.keySet())
		{	IMetaRepresentation meta = mLogicalEndpoints.get(key).mEndpointMeta;
			if (MetadataUtils.match((IEndpointMeta)meta, toResolve, aContext))
			{	result = key;
				break;
			}
		}
		return result;
	}

	public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception
	{	
		if(mAuto404!=null && aElementId.equals(mAuto404EndpointIdentifier))
		{	
			if(mAuto404.equals(""))
			{	INKFResponse resp=aContext.createResponseFrom("404 Resource not found");
				resp.setMimeType("text/plain");
				resp.setHeader("httpResponse:/code", 404);
				return;
			}
			else
			{	INKFRequest req=aContext.createRequestToEndpoint(mAuto404);
				req.addArgumentByValue("request", aContext.getThisRequest());
				INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
				INKFResponse resp=aContext.createResponseFrom(handlerResp);
				try
				{	handlerResp.getHeader("httpResponse:/code");
				}	
				catch(Exception e2)
				{	resp.setHeader("httpResponse:/code", 404);
				}
				return;
			}
		}
		LogicalEndpointData led=mLogicalEndpoints.get(aElementId);
		if(led.mMappings==null)
		{	//This is a delegated non-REST endpoint so just pass-through
			String targetID=mPassThroughEndpoints.get(aElementId);
			//INKFRequest req=aContext.createRequestToEndpoint(targetID);
			INKFRequest req=aContext.getThisRequest().getIssuableClone();
			
			aContext.createResponseFrom(aContext.issueRequestForResponse(req));
			return;
		}
		ArrayList<TargetEndpointData> mappings=led.mMappings;
		HttpServletRequest sreq;
		try
		{	sreq=aContext.source("httpRequest:/advanced/HttpServletRequest", HttpServletRequest.class);
		}
		catch(Exception e)
		{	throw new NKFException("RESTOverlay cannot find HTTPBridge","RESTOverlay can only be called when HTTPBridge is in the request scope");		
		}
		//Need this a few places so get it here
		String method=sreq.getMethod();
		
		INKFRequest req=null;
		//Sorting arrays
		ArrayList<TargetEndpointData> candidates=new ArrayList(4);
		ArrayList matches=new ArrayList(4);
		ArrayList possibles=new ArrayList(4);
		IHDSNode accepts=aContext.source("httpRequest:/accept", IHDSNode.class);
		String requiredMime=null;
		boolean matchedWithoutAcceptsCompare=false;
		/*
		IHDSNodeList acclist=accepts.getNodes("/headers/Accept/entry");
		if(acclist.size()==1)
		{	requiredMime=(String)acclist.get(0).getFirstValue("value");
		}
		if(acclist.size()>1)
		{	String firstMime=(String)acclist.get(0).getFirstValue("value");
			if(firstMime.indexOf("*")==-1)
			{	requiredMime=firstMime;				
			}
		}
		*/
		
		//System.out.println(accepts);
		while(true)
		{	//Method Matching
			for(TargetEndpointData ted: mappings)
			{	if(ted.mMethod!=null && ted.mMethod.contains(method))
				{	matches.add(ted);
				}
				else if(!mStrict)
				{	if(ted.mMethod==null)
					{	possibles.add(ted);					
					}
				}
			}
			filterLists(candidates, matches, possibles);
			if(!mStrict && candidates.size()==1)
			{	matchedWithoutAcceptsCompare=true;
				break;
			}
			
			//Accept Matching
			IHDSNodeList nl=accepts.getNodes("/headers/Accept/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					int wildcard=type.indexOf('*');
					if(wildcard>=0)
					{	type=type.substring(0, wildcard);						
					}
					for(TargetEndpointData ted: candidates)
					{	if(ted.mProduces!=null)
						{	
							for(String key: ted.mProduces.keySet())
							{	if(key.startsWith(type) && q!=0.0f)
								{	matches.add(ted);
									requiredMime=key;
									break;
								}
							}
						}
						else if(!mStrict)
						{	possibles.add(ted);
						}
					}
				}
				filterLists(candidates, matches, possibles);
			}
			if(!mStrict && candidates.size()==1)
			{	break;
			}
			
			//Consumes Matching
			if(HttpMethod.POST.is(method) || HttpMethod.PUT.is(method) || method.equals("PATCH"))
			{	String contentType=sreq.getContentType();
				int separator=contentType.indexOf(";");
				if(separator>0)
				{	contentType=contentType.substring(0, separator);
				}
				for(TargetEndpointData ted: candidates)
				{	if(!mStrict && ted.mConsumes==null)
					{	possibles.add(ted);
					}
					else
					{	if(ted.mConsumes.contains(contentType))
						{	matches.add(ted);
							break;
						}
					}
				}
				filterLists(candidates, matches, possibles);
			}
			if(!mStrict && candidates.size()==1)
			{	break;
			}
			
			//Accept-Language  Matching
			nl=accepts.getNodes("/headers/Accept-Language/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					for(TargetEndpointData ted: candidates)
					{	if(ted.mLanguage!=null)
						{	if(ted.mLanguage.equals(type)  && q!=0.0f)
							{	matches.add(ted);
								break;
							}
						}
						else if(!mStrict)
						{	possibles.add(ted);
						}
					}
				}
				filterLists(candidates, matches, possibles);
			}
			if(candidates.size()>1)
			{	if(mStrict)
				{	//Ambiguous 406
					do406(aContext);
					return;
				}
			}
			if(candidates.size()>=1)
			{	break;
			}
			//Nothing found so do 406
			do406(aContext);
			return;
		}
		//Issue Request to Target Endpoint
		TargetEndpointData target=candidates.get(0);

		//Do content size check
		if(HttpMethod.POST.is(method) || HttpMethod.PUT.is(method) || method.equals("PATCH"))
		{	int contentLength=sreq.getContentLength();
			boolean toobig=false;
			if(target.mMaxContentLength!=-1)
			{	toobig=contentLength>target.mMaxContentLength  || contentLength==-1;
			}
			else if (mGlobalMaxContentLength!=-1)
			{	toobig=contentLength>mGlobalMaxContentLength || contentLength==-1;;
			}
			if(toobig)
			{	do413(aContext, contentLength==-1);
				return;
			}
		}

		//Move to SSL?
		if(target.mMustBeSSL && !sreq.getProtocol().equals("https"))
		{	INKFResponse resp=aContext.createResponseFrom("302 Redirect - you must use SSL for this service");
			resp.setMimeType("text/plain");
			resp.setHeader("httpResponse:/redirect", sreq.getRequestURL().toString().replaceFirst("http://", "https://"));
			return;
		}
		//Pre Process
		if(target.mPreTarget!=null)
		{	INKFRequest pre=aContext.createRequestToEndpoint(target.mPreTarget);
			pre.addArgumentByValue("request", aContext.getThisRequest());
			pre.setRepresentationClass(Boolean.class);
			Boolean preresp=(Boolean)aContext.issueRequest(pre);
			if(!preresp)
			{	aContext.logRaw(INKFRequestContext.LEVEL_FINE, aContext.getThisRequest().toString()+" did not pass pre process "+target.mPreTarget);
				return;
			}
		}
		//Prepare and Issue Request
		req=prepareRequest(aContext,target.mIdentifier, target.mWrapperTarget, target.mVerbTranslate, method);
		Object resp=null;
		try
		{	resp=aContext.issueRequestForResponse(req);
		}
		catch(Exception e)
		{	//Exception Handling
			if(target.mExceptionTarget!=null)
			{	doExceptionHandling(aContext, e, target.mExceptionTarget);			
			}
			else if(!mExceptionHandler.equals(""))
			{	doExceptionHandling(aContext, e, mExceptionHandler);
			}
			else
			{	throw e;				
			}
			return;
		}
		
		//Do we need to use Produces withTransform?
		if(matchedWithoutAcceptsCompare)
		{	//Accept Matching
			IHDSNodeList nl=accepts.getNodes("/headers/Accept/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					if(type.equals("*/*") && q>0.0f)
					{	//They will accept anything so stop looking
						requiredMime=null;
					}
					int wildcard=type.indexOf('*');
					if(wildcard>=0)
					{	type=type.substring(0, wildcard);						
					}
					if(target.mProduces!=null)
					{	for(String key: target.mProduces.keySet())
						{	if(key.startsWith(type) && q!=0.0f)
							{	requiredMime=key;
								break;
							}
						}
					}
					if(requiredMime!=null)
					{	break;				
					}
				}
			}
		}
		if(requiredMime!=null && target.mProduces!=null)
		{	INKFResponseReadOnly r=(INKFResponseReadOnly)resp;
			if(!r.getMimeType().equals(requiredMime))
			{	//Do the withTransform
				IHDSNode produces=target.mProduces.get(requiredMime);
				if(produces!=null)
				{	String withTransform=(String)produces.getFirstValue("@withTransform");
					if(withTransform!=null)
					{	req=aContext.createRequest(withTransform);
						req.addArgumentByValue("operand", r.getRepresentation());
						resp=aContext.issueRequestForResponse(req);
					}
				}
			}
		}
		
		//Post Process Compression - take care if verb translating method is POST or PATCH as could be inconsistent
		if(target.mCompress && (HttpMethod.GET.is(method) || HttpMethod.POST.is(method) || method.equals("PATCH") ) )
		{	IHDSNodeList nl=accepts.getNodes("/headers/Accept-Encoding/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					if(type.equals("gzip")  && q!=0.0f)
					{	IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
						INKFResponseReadOnly roresp=(INKFResponseReadOnly)resp;
						INKFRequest compressRequest=aContext.createRequest("active:httpGZIPCompress");
						compressRequest.setRequestScope(scope);
						compressRequest.addArgumentFromResponse("operand", roresp);
						INKFResponseReadOnly cresp=aContext.issueRequestForResponse(compressRequest);
						INKFResponse compressedresp=aContext.createResponseFrom(cresp);
						relayHeaders(roresp, compressedresp);
						compressedresp.setHeader("httpResponse:/header/Content-Encoding", "gzip");
						resp=compressedresp;
						break;
					}
					else if(type.equals("deflate") && q!=0.0f)
					{	IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
						INKFResponseReadOnly roresp=(INKFResponseReadOnly)resp;
						INKFRequest compressRequest=aContext.createRequest("active:httpDeflateCompress");
						compressRequest.setRequestScope(scope);
						compressRequest.addArgumentFromResponse("operand", roresp);
						INKFResponseReadOnly cresp=aContext.issueRequestForResponse(compressRequest);
						INKFResponse compressedresp=aContext.createResponseFrom(cresp);
						relayHeaders(roresp, compressedresp);
						compressedresp.setHeader("httpResponse:/header/Content-Encoding", "deflate");
						resp=compressedresp;
						break;
					}
				}
			}
		}
		//Etag
		/*
		MD5, MD2, SHA-1,SHA-256,SHA-384,SHA-512
		 */
		if(HttpMethod.GET.is(method) && target.mEtag!=null)
		{	INKFResponse finalResponse;
			Object finalRep;
			if(resp instanceof INKFResponse)
			{	finalResponse=(INKFResponse)resp;
				finalRep=finalResponse.getRepresentation();
			}
			else
			{	finalResponse=aContext.createResponseFrom(resp);
				finalRep=((INKFResponseReadOnly)resp).getRepresentation();
			}
			if(finalRep instanceof IReadableBinaryStreamRepresentation)
			{	
				INKFRequest etagReq=aContext.createRequest("active:httpEtag");
				etagReq.addArgumentByValue("operand", finalRep);
				etagReq.addArgumentByValue("operator", target.mEtag);
				IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
				etagReq.setRequestScope(scope);
				String eTag=(String)aContext.issueRequest(etagReq);
				finalResponse.setHeader("httpResponse:/header/Etag", eTag);
				resp=finalResponse;
			}
		}
		//Return the response
		aContext.createResponseFrom(resp);
	}
	
	private void doExceptionHandling(INKFRequestContext aContext, Exception e, String handler) throws Exception
	{	INKFRequest req=aContext.createRequestToEndpoint(handler);
		req.addArgumentByValue("request", aContext.getThisRequest());
		req.addArgumentByValue("exception", new RESTOverlayWrappedException(e));
		INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
		INKFResponse resp=aContext.createResponseFrom(handlerResp);
		try
		{	handlerResp.getHeader("httpResponse:/code");
		}	
		catch(Exception e2)
		{	resp.setHeader("httpResponse:/code", 500);
		}
		return;
	}
	
	private void do406(INKFRequestContext aContext) throws Exception
	{	if(mAuto406.equals(""))
		{	String message;
			if(mStrict)
			{ message="406 Resource found but more than one possible endpoint so not acceptable in strict mode";
			}
			else
			{	message="406 Resource found but not acceptable";				
			}
			INKFResponse resp=aContext.createResponseFrom(message);
			resp.setMimeType("text/plain");
			resp.setHeader("httpResponse:/code", 406);
			return;
		}
		else
		{	INKFRequest req=aContext.createRequestToEndpoint(mAuto406);
			req.addArgumentByValue("request", aContext.getThisRequest());
			INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
			INKFResponse resp=aContext.createResponseFrom(handlerResp);
			resp.setHeader("httpResponse:/code", 406);
			return;
		}
	}
	
	private void do413(INKFRequestContext aContext, boolean notReported) throws Exception
	{	String message;
		if(notReported)
		{	message="413 Content-Length header not reported. Must be set to check acceptable maximum permitted size";
		}
		else
		{	message="413 Content size exceeds maximum permitted";
		}
		INKFResponse resp=aContext.createResponseFrom(message);
		resp.setMimeType("text/plain");
		resp.setHeader("httpResponse:/code", 413);
		return;
	}
	
	private void relayHeaders(INKFResponseReadOnly oldresp, INKFResponse newresp)
	{	IRequestResponseFields fields=oldresp.getHeaders();
		for(int i=0; i< fields.size(); i++)
		{	String key=fields.getKey(i);
			Object value=fields.getValue(i);
			newresp.setHeader(key, value);
		}
	}
	
	private IRequestScopeLevel getWormholeScopeToHTTPTools(INKFRequestContext aContext) throws Exception
	{	IKernel kernel=aContext.getKernelContext().getKernel();
		String spaceId="urn:org:netkernel:tpt:http";
		ISpace targetSpace=kernel.getSpace(new SimpleIdentifierImpl(spaceId), null, null);
		if (targetSpace==null)
		{	throw aContext.createFormattedException("EX_SPACE_NOT_FOUND", "MSG_SPACE_NOT_FOUND", null, null, spaceId);
		}
		return RequestScopeLevelImpl.createOrphanedRootScopeLevel(targetSpace, aContext.getKernelContext().getRequestScope());
	}
	
	private void filterLists(ArrayList candidates, ArrayList matches, ArrayList possibles)
	{	candidates.clear();
		for(Object o: matches)
		{	candidates.add(o);
		}
		for(Object o: possibles)
		{	candidates.add(o);
		}
		matches.clear();
		possibles.clear();
	}
	
	private INKFRequest prepareRequest(INKFRequestContext aContext, String aTargetId, String aWrapperTargetId, Map<String, Integer> aVerbTranslate, String method) throws Exception
	{	INKFRequest req=aContext.createRequestToEndpoint(aTargetId);
		INKFRequestReadOnly logicalReq=aContext.getThisRequest();
		for(int i=0; i<logicalReq.getArgumentCount(); i++)
		{	String argName=logicalReq.getArgumentName(i);
			if(argName.equals("argType") || argName.equals("scheme"))
			{	break;
			}
			String argValue=logicalReq.getArgumentValue(i);
			req.addArgument(argName, argValue);
		}
		//Wrap the request with the wrapper
		if(aWrapperTargetId!=null)
		{	if(aVerbTranslate!=null)
			{	throw new Exception("Request Wrapping is incompatible with Verb Translation - either don't wrap or don't translate");			
			}
			INKFRequest wrappedReq=req;
			req=aContext.createRequestToEndpoint(aWrapperTargetId);
			WrappedRequestValueSpace vspace=new WrappedRequestValueSpace(wrappedReq);
			vspace.onCommissionSpace(aContext.getKernelContext().getKernel());
			IRequestScopeLevel scope=aContext.getKernelContext().getRequestScope();
			scope=RequestScopeLevelImpl.createDurableScopeLevel(vspace, scope);
			req.setRequestScope(scope);
			req.addArgument("operand", vspace.PBV_IDENTIFIER);
		}
		//VerbTranslation
		if(aVerbTranslate!=null)
		{	Integer verb=aVerbTranslate.get(method);
			if(verb==null)
			{	verb=INKFRequestReadOnly.VERB_SOURCE;				
			}
			if(verb==INKFRequestReadOnly.VERB_SINK  || verb==INKFRequestReadOnly.VERB_NEW)
			{	INKFResponseReadOnly bodyResp=aContext.sourceForResponse("httpRequest:/body");
				req.addPrimaryArgumentFromResponse(bodyResp);
			}
			req.setVerb(verb);
		}
		req.setRepresentationClass(aContext.getThisRequest().getRepresentationClass());
		return req;
	}
		
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	return new ISpace[] { mDelegateSpace };
	}

	private void initialize(INKFRequestContext aContext, RESTOverlay aOverlay) throws Exception
	{	ISpaceMeta sm = mDelegateSpaceMeta;
		ISpaceElements sme = sm.getElements();
		Map<String, String> uniqueGrammars=new HashMap();
		boolean errors=false;
		StringBuilder errorBuilder=new StringBuilder();
		for(int i=0; i<sme.size(); i++)
		{	IIdentifier id=sme.getIdentifier(i);
			if(id!=null)
			{	String targetEPID=id.toString();
				IMetaRepresentation meta=aContext.meta(targetEPID);
				IRequestResponseFields state = meta.getAdditionalFields();
				Object value = state.getValue("meta");
				if (value instanceof Element)
				{	IHDSNode n = HDSFactory.parseDOM((Element) value);
					n=n.getFirstNode("/meta/rest");
					if(n!=null)
					{	String simple=(String)n.getFirstValue("simple");
						if(simple==null)
						{	simple="";						
						}
						String grammar=mBasePath+simple;
						String aIdentifier=constructIdentifier(grammar);
						LogicalEndpointData led=mLogicalEndpoints.get(aIdentifier);
						if(led==null)
						{	led=new LogicalEndpointData(aIdentifier, grammar, aContext.getKernelContext().getKernel().getLogger(), state);
						}
						//Check that target endpoint grammar is unique and log a warning if it is not
						if(meta instanceof IEndpointMeta)
						{	IEndpointMeta epm=(IEndpointMeta)meta;
							String gs=epm.getIdentifierGrammar().toString();
							String existingGrammar=uniqueGrammars.get(gs);
							if(existingGrammar!=null)
							{	aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "WARNING_RESTOVERLAY_AMBIGUOUSGRAMMAR", existingGrammar, targetEPID);
								errors=true;
								errorBuilder.append("Ambiguous Grammar - the target endpoint ["+targetEPID+"]  has a non-unique grammar ["+existingGrammar+"]\n");
							}
							else
							{	uniqueGrammars.put(gs, gs);								
							}
						}
						led.mMappings.add(new TargetEndpointData(targetEPID, n));
						mLogicalEndpoints.put(led.mIdentifier, led);
					}
				}
				else
				{	//Transparently relay through any non-RESTOverlay managed endpoint
					if(meta instanceof IEndpointMeta)
					{	//String proxyID=constructIdentifier(targetEPID);
						mLogicalEndpoints.put(targetEPID, new LogicalEndpointData(targetEPID, meta));
						mPassThroughEndpoints.put(targetEPID, targetEPID);
					}
				}
			}
			if(errors)
			{	aOverlay.setInitialisationProblem(errorBuilder.toString());
			}
		}
		// Construct space elements with a tuple list {IIdentifier aIndentifier, ISpace aDelegatedSpace, IEndpoint aPhysicalEndpoint}
		int tlSize=0;
		if(mLogicalEndpoints.size()!=0)
		{	tlSize=mLogicalEndpoints.size();			
		}
		else if(mAuto404!=null)
		{	tlSize=1;			
		}
		TupleList tl=new TupleList(3, tlSize);
		for(String key: mLogicalEndpoints.keySet())
		{		tl.put(new SimpleIdentifierImpl(key), null, mRESTOverlay);
		}
		if(mAuto404!=null)
		{		String grammar=mBasePath+"{404}";
				mAuto404EndpointIdentifier=constructIdentifier(grammar);
				LogicalEndpointData led=new LogicalEndpointData(mAuto404EndpointIdentifier, grammar, aContext.getKernelContext().getKernel().getLogger(), RequestResponseFieldsImpl.EMPTY);
				mLogicalEndpoints.put(led.mIdentifier, led);
				tl.put(new SimpleIdentifierImpl(mAuto404EndpointIdentifier), null, mRESTOverlay);
		}
		
		mSpaceElements=new SpaceElementsImpl(tl);
	}
	
	private class LogicalEndpointData
	{	public String mIdentifier;
		public ArrayList<TargetEndpointData> mMappings;
		public IMetaRepresentation mEndpointMeta;
		
		public LogicalEndpointData(String aIdentifier, IMetaRepresentation aEndpointMeta)
		{	mEndpointMeta=aEndpointMeta;
			mIdentifier=aIdentifier;
		}
		
		public LogicalEndpointData(String aIdentifier, String aGrammar, IMessages aMessages, IRequestResponseFields aFields) throws Exception
		{	mIdentifier=aIdentifier;
			constructMetadata(aGrammar, aMessages, aFields);
			mMappings=new ArrayList();
		}
		
		private void constructMetadata(String aGrammar, IMessages aMessages, IRequestResponseFields aFields) throws Exception
		{	String grammarImpl="res:"+aGrammar;
			String grammar="<grammar><simple>"+grammarImpl+"</simple></grammar>";
			Document gd=XMLUtils.parse(new ByteArrayInputStream(grammar.getBytes()));
			//IIdentifierGrammar g=BNFGrammarFactory.parse(gd.getDocumentElement(), aMessages);
			EndpointMetaBuilder epm=new EndpointMetaBuilder();
			MetadataUtils.parseGrammar(gd.getDocumentElement(), epm, aMessages);
			IEndpointMeta meta=epm.getMeta(aMessages);
			ClonableEndpointMetaImpl meta2=new ClonableEndpointMetaImpl(
					grammarImpl,
					"Auto-generated RESTOverlay logical endpoint",
					null,
					null,
					INKFRequestReadOnly.VERB_SOURCE,
					meta.getIdentifierGrammar(),
					EndpointMetaImpl.NO_ARGS,
					EndpointMetaImpl.NO_RESPONSES,
					EndpointMetaImpl.NO_EXCEPTIONS
				);
			meta2.setAdditionalFields(aFields);
			mEndpointMeta=meta2;
		}
	}
	
	private class TargetEndpointData
	{	public String mIdentifier;
		public IHDSNode mMeta;
		public String mMethod;
		public Map<String, IHDSNode> mProduces;
		public String mConsumes;
		public String mLanguage;
		public boolean mCompress;
		public boolean mMustBeSSL;
		public String mExceptionTarget;
		public String mPreTarget;
		public String mWrapperTarget;
		public String mEtag;
		public int mMaxContentLength=-1;
		public Map<String, Integer> mVerbTranslate;
		
		public TargetEndpointData(String aIdentifier, IHDSNode aMeta)
		{	mIdentifier=aIdentifier;
			mMeta=aMeta;
			analyseMeta();
		}
		
		private void analyseMeta()
		{	String method=(String)mMeta.getFirstValue("method");
			if(method!=null)
			{	method=method.toUpperCase();				
			}
			mMethod=method;
			//mProduces=(String)mMeta.getFirstValue("produces");
			IHDSNodeList nl=mMeta.getNodes("produces");
			if(nl.size()>0)
			{	mProduces=new HashMap(nl.size());
				for(IHDSNode n: nl)
				{	String v=(String)n.getValue();
					mProduces.put(v, n);
				}
			}
			mConsumes=(String)mMeta.getFirstValue("consumes");
			mLanguage=(String)mMeta.getFirstValue("language");
			mCompress=mMeta.getFirstNode("compress")!=null;
			mMustBeSSL=mMeta.getFirstNode("mustBeSSL")!=null;
			mExceptionTarget=(String)mMeta.getFirstValue("exceptionTarget");
			mPreTarget=(String)mMeta.getFirstValue("preTarget");
			mWrapperTarget=(String)mMeta.getFirstValue("wrapperTarget");
			mEtag=(String)mMeta.getFirstValue("Etag");
			if(mEtag!=null)
			{	mEtag=mEtag.toUpperCase();				
			}
			String maxContentLength=(String)mMeta.getFirstValue("maxContentLength");
			if(maxContentLength!=null)
			{	mMaxContentLength=Integer.parseInt(maxContentLength);
			}
			boolean doVerbTranslate=mMeta.getFirstNode("verbTranslate")!=null;
			if(doVerbTranslate)
			{	mVerbTranslate=new HashMap();
				mVerbTranslate.put("GET", INKFRequestReadOnly.VERB_SOURCE);
				mVerbTranslate.put("POST", INKFRequestReadOnly.VERB_SINK);
				mVerbTranslate.put("PUT", INKFRequestReadOnly.VERB_SINK);
				mVerbTranslate.put("DELETE", INKFRequestReadOnly.VERB_DELETE);
				String verbTranslate=(String)mMeta.getFirstValue("verbTranslate");
				if(verbTranslate!=null)
				{	Class c=INKFRequestReadOnly.class;
					verbTranslate=verbTranslate.toUpperCase();
					String[] s1=verbTranslate.split(",");
					for(int i=0; i<s1.length; i++)
					{	String s=s1[i].trim();
						String[] s2=s.split(":");
						String m=s2[0];
						String v=s2[1];
						try
						{	Field f=c.getField("VERB_"+v);
							mVerbTranslate.put(m, f.getInt(null));
						}
						catch(Exception e)
						{	//Do nothing							
						}
						
					}
				}
			}
		}
	}
	
	private class ClonableEndpointMetaImpl extends MetaRepresentationImpl implements IEndpointMeta
	{
		private final int mVerbs;
		private final IIdentifierGrammar mGrammar;
		private final IEndpointArgumentMeta[] mParts;
		private final Class[] mResponseClasses;
		private final IMetaRepresentation[] mExceptions;
		
		public ClonableEndpointMetaImpl(String aName, String aDescription, IIdentifier aIcon, IIdentifier aDoc,
				int aVerbs, IIdentifierGrammar aGrammar, IEndpointArgumentMeta[] aParts, Class[] aResponseClasses, IMetaRepresentation[] aExceptions)
		{	super(aName,aDescription,aIcon,aDoc);
			mVerbs=aVerbs;
			mGrammar=aGrammar;
			mParts=aParts;
			mResponseClasses=aResponseClasses;
			mExceptions = aExceptions;
		}

		public int getSupportedVerbs()
		{	return mVerbs;
		}

		public IIdentifierGrammar getIdentifierGrammar()
		{	return mGrammar;
		}

		public IEndpointArgumentMeta[] getArguments()
		{	return mParts;
		}

		public Class[] getResponseClasses()
		{	return mResponseClasses;
		}

		public IMetaRepresentation[] getExceptions()
		{	return mExceptions;
		}
		
		public void setAdditionalFields(IRequestResponseFields aFields)
		{	super.setAdditionalFields(aFields);
		}
	}
	
	public String constructIdentifier(String aId)
	{	return aId.replaceAll("/", "_").replaceAll("\\?", "_");
		//return Utils.generateUniqueEndpointIdentifier(mRESTOverlay.getOwningSpace(), mRESTOverlay, aId);
	}
}
