package org.netkernel.http.transport;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.http.HttpHeader;
import org.eclipse.jetty.http.HttpMethod;
import org.netkernel.container.IKernel;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.http.transport.representation.NKWebSocket;
import org.netkernel.http.transport.representation.ParameterUploadAspect;
import org.netkernel.http.util.AcceptHeaderParser;
import org.netkernel.http.util.MultiPartRequest;
import org.netkernel.http.util.SecurityUtils;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFTransportContextImpl;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.CheapMap;
import org.netkernel.layer0.util.RegexRewriterRule;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestScopeLevelImpl;


public class HTTPRequestSpace2Wrapper extends StandardSpace
{
	public static final String HTTP_REQUEST_SCHEME="httpRequest:/";
	public static final String HTTP_RESPONSE_SCHEME="httpResponse:/";
	public static final String WEBSOCKET_RESPONSE_SCHEME="wsResponse:/";
	public static final String HTTP_PARAM_SET="params";
	public static final String HTTP_PARAM_SET_HTML_ENCODED="paramsHTMLEncoded";
	public static final String HTTP_PARAM_SPACE="param/";
	public static final String HTTP_PARAM_SPACE_HTML_ENCODED="paramHTMLEncoded/";
	public static final String HTTP_POST_PARAM_SPACE="postParam/";
	public static final String HTTP_POST_PARAM_SPACE2="postparam/";
	public static final String HTTP_QUERY_STRING="query";
	public static final String HTTP_QUERY_STRING2="url/query";
	public static final String HTTP_URL="url";
	public static final String HTTP_BODY="body";
	public static final String HTTP_UPLOAD="upload/";
	public static final String HTTP_METHOD="method";
	public static final String HTTP_ACCEPT="accept";
	public static final String HTTP_HEADER="header/";
	public static final String HTTP_HEADER_HTML_ENCODED="headerHTMLEncoded/";
	public static final String HTTP_HEADERS="headers";
	public static final String HTTP_HEADERS_HTML_ENCODED="headersHTMLEncoded";
	public static final String HTTP_RESPONSE_CODE="code";
	public static final String HTTP_COOKIE="cookie/";
	public static final String HTTP_COOKIE_SINK="cookie";
	public static final String HTTP_REMOTEHOST="remote-host";
	public static final String HTTP_REDIRECT="redirect";
	public static final String HTTP_LOW_LEVEL_REQUEST="advanced/HttpServletRequest";
	public static final String HTTP_LOW_LEVEL_RESPONSE="advanced/HttpServletResponse";
	//New resources
	public static final String HTTP_PATH="url/path";
	public static final String HTTP_FULLURL="url/full";
	public static final String HTTP_URL_RELATIVE="url/relative";
	public static final String HTTP_CONTENT_TYPE="content-type";
	public static final String HTTP_CONTENT_TYPE_PARAMS="content-type/params";
	public static final String HTTP_ACCEPT_PREFERRED="accept/preferred";
	public static final String HTTP_ACCEPT_PREFERRED_PATH="accept/preferred/";
	public static final String HTTP_ACCEPTABLE_PATH="acceptable/";
	//Matrix URI
	public static final String HTTP_MATRIXPARAM_SET="matrixparams";
	public static final String HTTP_MATRIXPARAM_SET_HTML_ENCODED="matrixparamsHTMLEncoded";
	public static final String HTTP_MATRIXPARAM_SPACE="matrixparam/";
	public static final String HTTP_MATRIXPARAM_SPACE_HTML_ENCODED="matrixparamHTMLEncoded/";
	
	public static final String HTTP_QUERYPARAM_SPACE="queryparam/";
	public static final String HTTP_QUERYPARAM_SPACE_HTML_ENCODED="queryparamHTMLEncoded/";
	
	public static final String HTTP_COLLECTION_SET="collection";
	public static final String HTTP_COLLECTION_SET_HTML_ENCODED="collectionHTMLEncoded";
	
	private final HttpServletRequest mRequest;
	private HTTPBridgeOverlayConfig mConfig;
	private final NKWebSocket mWS;
	private final INKFRequestContext mTransportContext;
	private final String mURL;
	private final int mHash;
	private final String mMethod;
	private static IHDSNode sFailedParam;
	static
	{	HDSBuilder b=new HDSBuilder();
		sFailedParam=b.getRoot();
	}
	
	/** map of String -> IReadableBinaryStreamRepresentation */
	private Map mUploads;
	/** content of request body (post or put of data) */
	private IReadableBinaryStreamRepresentation mContent;
	/** parameters */
	private IHDSNode mParam;
	private IHDSNode mMatrixParam;
	private int mParamHash;
	private HTTPRequestResponseRepresentation mRequestResponse;
	
	//Response Resources
	private Map mResponseCookies=new CheapMap(2);
	private Map mResponseHeaders=new CheapMap(2);
	private int mResponseCode=200;
	private String mResponseRedirect;	
	
	public HTTPRequestSpace2Wrapper(HTTPBridgeOverlay aBridge, HTTPRequestSpace2 aWrappedSpace, HTTPRequestResponseRepresentation aRequestResponse, HTTPBridgeOverlayConfig aConfig, NKWebSocket aWS)
	{	super(aBridge.getOwningSpace(),aBridge.getOwningSpace().getOwningModule(),null);
		mRequestResponse=aRequestResponse;
		mRequest=aRequestResponse.getHttpRequest();
		mConfig=aConfig;
		mWS=aWS;
		setMappings(aWrappedSpace.getEndpoints());
		setName("HTTP Request Response Space");
		setState(ISpaceMeta.STATE_PROTOSPACE, aWrappedSpace );
		
		mURL=Utils.decode(mRequest.getRequestURL().toString());
		mHash=mURL.hashCode();
		mMethod=mRequest.getMethod();		
		
		if (mWS!=null)
		{	IRequestScopeLevel injectionScope = RequestScopeLevelImpl.createRootScopeLevel(this);
			injectionScope=RequestScopeLevelImpl.createNonDurableScopeLevel(aBridge.getDelegateSpace(), injectionScope);
			IKernel k=aWrappedSpace.getKernel();
			mTransportContext = new NKFTransportContextImpl(k, this, this, null, injectionScope);
			mWS.setTransportSpace(this);
		}
		else
		{	mTransportContext=null;
		}
	}
	
	public void releaseReferences()
	{	super.releaseReferences();
		if (mWS==null)
		{	mConfig=null;
		}
	}
	
	public RegexRewriterRule[] getRewrtieRules()
	{	return mConfig.getRewriteRules();		
	}
	
	public INKFRequestContext getTransportContext()
	{	return mTransportContext;
	}
	
	public void disconnectWS()
	{	if (mConfig!=null)
		{	mConfig.remove(this);
			mConfig=null;
		}
	}
	
	public void onRequestFromEndpoint(INKFRequestContext aContext) throws Exception
	{	
		Object response=null;
		String contentType="text/plain";
		int verb=aContext.getThisRequest().getVerb();
		String identifier=aContext.getThisRequest().getIdentifier();
	
		if(identifier.startsWith(HTTP_REQUEST_SCHEME))
		{	if(verb!=INKFRequestReadOnly.VERB_SOURCE && verb!=INKFRequestReadOnly.VERB_EXISTS)
			{	throw new NKFException("httpRequest is SOURCE,EXISTS only");				
			}
			boolean fullAccess=true;
			String type = identifier.substring(HTTP_REQUEST_SCHEME.length());
			if (type.equals(HTTP_PARAM_SET_HTML_ENCODED))
			{	response = getParamSetHTMLEncoded();
				contentType=null;
				fullAccess=false;
			}
			else if (type.equals(HTTP_PARAM_SET))
			{	response = getParamSet();
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_PARAM_SPACE_HTML_ENCODED))
			{	response = getParamSpaceHTMLEncoded(type);
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_PARAM_SPACE))
			{	response = getParamSpace(type);
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_QUERYPARAM_SPACE))
			{	ensureParamsProcessed();
				String param = type.substring(HTTP_QUERYPARAM_SPACE.length());
				String queryString=mRequest.getQueryString();
				if(queryString!=null && queryString.contains(param+"="))
				{	response=mParam.getFirstValue(param);	
				}
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_QUERYPARAM_SPACE_HTML_ENCODED))
			{	ensureParamsProcessed();
				String param = type.substring(HTTP_QUERYPARAM_SPACE_HTML_ENCODED.length());
				String queryString=mRequest.getQueryString();
				if(queryString!=null && queryString.contains(param+"="))
				{	response=mParam.getFirstValue(param);
					if(response!=null)
					{	response=SecurityUtils.HTMLEncode((String)response);
					}
				}
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_POST_PARAM_SPACE2) || type.startsWith(HTTP_POST_PARAM_SPACE))
			{	if(!HttpMethod.POST.is(mMethod))
				{	throw new NKFException("http:/request/postparam/ can only be SOURCEed when HTTP request is POST");				
				}
				ensureParamsProcessed();
				String param = type.substring(HTTP_POST_PARAM_SPACE.length());
				String queryString=mRequest.getQueryString();
				if(queryString!=null && queryString.contains(param+"="))
				{	throw new NKFException("A query parameter with the same name has been supplied. Possible security violation. Recommend rejecting the request");
				}
				response=mParam.getFirstValue(param);
				contentType=null;
				fullAccess=false;
			}
			else if (type.equals(HTTP_BODY))
			{	ensureParamsProcessed();
				if (mContent!=null)
				{	response= mContent;
					contentType=mRequest.getContentType();
				}
			}
			else if (type.equals(HTTP_URL))
			{	response = getURL();
				fullAccess=false;
			}
			else if (type.equals(HTTP_QUERY_STRING) || type.equals(HTTP_QUERY_STRING2))
			{	response = getURLQuery();
				fullAccess=false;
			}
			else if (type.equals(HTTP_PATH))
			{	response = getURLPath();
				fullAccess=false;
			}
			else if (type.equals(HTTP_FULLURL))
			{	response = getURLFull();
				fullAccess=false;
			}
			else if (type.equals(HTTP_URL_RELATIVE))
			{	response = getURLRelative();
				fullAccess=false;
			}
			else if (type.equals(HTTP_COLLECTION_SET))
			{	response = getCollection();
				fullAccess=false;
			}
			else if (type.equals(HTTP_COLLECTION_SET_HTML_ENCODED))
			{	response = getCollectionHTMLEncoded();
				fullAccess=false;
			}
			else if (type.equals(HTTP_CONTENT_TYPE))
			{	String ct=mRequest.getHeader(HttpHeader.CONTENT_TYPE.asString());
				if(ct!=null)
				{	String[] tokens = ct.split(";");
					response= tokens.length>=1 ? tokens[0] : ct;
				}
			}
			else if (type.equals(HTTP_CONTENT_TYPE_PARAMS))
			{	String ct=mRequest.getHeader(HttpHeader.CONTENT_TYPE.asString());
				if(ct!=null)
				{	int i=ct.indexOf(";");
					if(i>0)
					{	response=ct.substring(i+1).trim();					
					}
					else
					{	response="";				
					}
				}
			}
			else if (type.startsWith(HTTP_UPLOAD))
			{	ensureParamsProcessed();
				String upload = type.substring(HTTP_UPLOAD.length());
				if(mUploads==null)
				{	throw new Exception("No file uploads - check form enctype attribute is \"multipart/form-data\" ");					
				}
				response = mUploads.get(upload);
				contentType=null;
			}
			else if (type.equals(HTTP_METHOD))
			{	response = getMethod();
				fullAccess=false;
			}
			else if (type.equals(HTTP_ACCEPT))
			{	response=AcceptHeaderParser.getAcceptHeaderState(mRequest);
			}
			else if (type.equals(HTTP_ACCEPT_PREFERRED))
			{	
				IHDSNode hds=AcceptHeaderParser.getAcceptHeaderState(mRequest);
				String pref=(String)hds.getFirstValue("/headers/Accept/entry[1]/value");
				response=pref;
			}
			else if (type.startsWith(HTTP_ACCEPT_PREFERRED_PATH))
			{	String param = type.substring(HTTP_ACCEPT_PREFERRED_PATH.length());
				String[] tokens = param.split(",");
				IHDSNode hds=AcceptHeaderParser.getAcceptHeaderState(mRequest);
				response = null;
				for(int i=0; i<tokens.length; i++)
				{	String t=tokens[i].trim();
					String tcompare=t.toUpperCase();
					IHDSNodeList nl=hds.getNodes("/headers/Accept/entry");
					for( IHDSNode n: nl)
					{	String v=((String)n.getFirstValue("value")).toUpperCase();
						if(tcompare.equals(v))
						{	response=t;
							break;
						}
					}
				}
			}
			else if (type.startsWith(HTTP_ACCEPTABLE_PATH))
			{	String param = type.substring(HTTP_ACCEPTABLE_PATH.length()).toUpperCase();
				IHDSNode hds=AcceptHeaderParser.getAcceptHeaderState(mRequest);
				response = false;
				IHDSNodeList nl=hds.getNodes("/headers/Accept/entry");
				for( IHDSNode n: nl)
				{	String v=((String)n.getFirstValue("value")).toUpperCase();
					if(param.equals(v))
					{	response=true;
						break;
					}
					else if(v.contains("*"))
					{	int i=v.indexOf("*");
						v=v.substring(0, i);
						if(param.contains(v))
						{	response=true;
							break;						
						}
					}
				}
			}
			else if (type.equals(HTTP_MATRIXPARAM_SET_HTML_ENCODED))
			{	response = getMatrixParamsHTMLEncoded();
				contentType=null;
				fullAccess=false;
			}
			else if (type.equals(HTTP_MATRIXPARAM_SET))
			{	response = getMatrixParams();
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_MATRIXPARAM_SPACE_HTML_ENCODED))
			{	ensureMatrixParamsProcessed();
				String param = type.substring(HTTP_MATRIXPARAM_SPACE_HTML_ENCODED.length());
				response=SecurityUtils.HTMLEncode((String)mMatrixParam.getFirstValue(param));
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_MATRIXPARAM_SPACE))
			{	ensureMatrixParamsProcessed();
				String param = type.substring(HTTP_MATRIXPARAM_SPACE.length());
				response=mMatrixParam.getFirstValue(param);
				contentType=null;
				fullAccess=false;
			}
			else if (type.startsWith(HTTP_HEADER_HTML_ENCODED))
			{	String field = type.substring(HTTP_HEADER_HTML_ENCODED.length());
				response = SecurityUtils.HTMLEncode(mRequest.getHeader(field));
			}
			else if (type.startsWith(HTTP_HEADER))
			{	String field = type.substring(HTTP_HEADER.length());
				response = mRequest.getHeader(field);
			}
			else if (type.startsWith(HTTP_HEADERS_HTML_ENCODED))
			{	HDSBuilder b=new HDSBuilder();
				Enumeration e=mRequest.getHeaderNames();
				while(e.hasMoreElements())
				{	String name=(String)e.nextElement();
					Enumeration f=mRequest.getHeaders(name);
					while(f.hasMoreElements())
					{	String value=SecurityUtils.HTMLEncode((String)f.nextElement());
						b.addNode(name, value);
					}
				}
				response = b.getRoot();
			}
			else if (type.startsWith(HTTP_HEADERS))
			{	HDSBuilder b=new HDSBuilder();
				Enumeration e=mRequest.getHeaderNames();
				while(e.hasMoreElements())
				{	String name=(String)e.nextElement();
					Enumeration f=mRequest.getHeaders(name);
					while(f.hasMoreElements())
					{	String value=(String)f.nextElement();
						b.addNode(name, value);						
					}
				}
				response = b.getRoot();
			}
			else if (type.startsWith(HTTP_COOKIE))
			{	String name = type.substring(HTTP_COOKIE.length());
				Cookie[] cookies = mRequest.getCookies();
				if(cookies!=null)
				{	for(int i=0; i<cookies.length; i++)
					{	if (cookies[i].getName().equals(name))
						{	response = cookies[i];
							break;
						}
					}
				}
			}
			else if (type.equals(HTTP_REMOTEHOST))
			{	response=mRequest.getRemoteHost();
			}
			else if (type.equals(HTTP_LOW_LEVEL_REQUEST))
			{	response=mRequest;
			}
			else if (type.equals(HTTP_LOW_LEVEL_RESPONSE))
			{	response=mRequestResponse.getHttpResponse();
				// stop bridge from handling response
				mRequestResponse.needsHandling();
			}
			else
			{	INKFRequestReadOnly req = aContext.getThisRequest();
				String verb2=RequestFactory.verbString(req.getVerb());
				NKFException e=aContext.createFormattedException("EX_BAD_HTTP_RESOURCE", "MSG_BAD_HTTP_RESOURCE", null,null,verb2,req.getIdentifier());
				throw e;
			}
			
			if (aContext.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_EXISTS)
			{	boolean exists = response!=null;
				aContext.createResponseFrom( exists ? Boolean.TRUE : Boolean.FALSE );
			}
			else
			{	INKFResponse nkfResp=aContext.createResponseFrom(response);
				nkfResp.setExpiry(INKFResponse.EXPIRY_NEVER);
				if (contentType!=null)
				{	nkfResp.setMimeType(contentType);
				}
			}
			
			if (fullAccess)
			{	mAdvancedAccess=true;
			}
			else
			{	mBasicAccess=true;
			}
			
		}
		else if(identifier.startsWith(WEBSOCKET_RESPONSE_SCHEME))
		{	String type = identifier.substring(WEBSOCKET_RESPONSE_SCHEME.length());
			if(verb!=INKFRequestReadOnly.VERB_SINK)
			{	throw new NKFException("wsResponse is SINK only");
			}
			Object value=aContext.getThisRequest().getPrimary();
			innerProcessWSResponseSink(type,value,aContext);
		}
		else
		{	String type = identifier.substring(HTTP_RESPONSE_SCHEME.length());
			if(verb!=INKFRequestReadOnly.VERB_SINK)
			{	throw new NKFException("httpResponse is SINK only");
			}
			Object value=aContext.getThisRequest().getPrimary();
			innerProcessResponseSink(type,value,aContext);
		}
	}

	private Object getCollection() throws Exception
	{	
		HDSBuilder b=new HDSBuilder();
		b.pushNode("hds");
		b.addNode("url", this.getURL());
		b.addNode("urlPath", this.getURLPath());
		b.addNode("urlQuery", this.getURLQuery());
		b.addNode("urlRelative", this.getURLRelative());
		b.addNode("method", this.getMethod());
		b.pushNode("params");
		b.importChildren((IHDSNode)this.getQueryPostParams());
		b.popNode();
		b.pushNode("matrixparams");
		b.importChildren((IHDSNode)this.getMatrixParams());
		b.popNode();
		
		Object response;
		response = b.getRoot();
		return response;
	}
	
	private Object getCollectionHTMLEncoded() throws Exception
	{	
		HDSBuilder b=new HDSBuilder();
		b.pushNode("hds");
		b.addNode("@encoding","HTML-Encoded");
		b.addNode("url", SecurityUtils.HTMLEncode((String)this.getURL()));
		b.addNode("urlPath", SecurityUtils.HTMLEncode((String)this.getURLPath()));
		b.addNode("urlQuery", SecurityUtils.HTMLEncode((String)this.getURLQuery()));
		b.addNode("urlRelative", SecurityUtils.HTMLEncode((String)this.getURLRelative()));
		b.addNode("method", SecurityUtils.HTMLEncode((String)this.getMethod()));
		b.pushNode("params");
		b.importChildren((IHDSNode)getQueryPostParamsHTMLEncoded());
		b.popNode();
		b.pushNode("matrixparams");
		b.importChildren((IHDSNode)getMatrixParamsHTMLEncoded());
		b.popNode();
		
		Object response;
		response = b.getRoot();
		return response;
	}

	private Object getMatrixParams() throws IOException
	{
		Object response;
		ensureMatrixParamsProcessed();
			response=mMatrixParam;
		return response;
	}
	
	private Object getQueryPostParams() throws IOException
	{
		Object response;
		ensureParamsProcessed();
		response=mParam;
		return response;
	}
	
	private Object getQueryPostParamsHTMLEncoded() throws IOException
	{
		Object response;
		ensureParamsProcessed();
		HDSBuilder b=new HDSBuilder();
		IHDSNodeList nl=mParam.getNodes("/*");
		for(IHDSNode n: nl)
		{	b.addNode(n.getName(), SecurityUtils.HTMLEncode((String)n.getValue()));					
		}
		response=b.getRoot();
		return response;
	}

	private Object getMatrixParamsHTMLEncoded() throws IOException
	{
		Object response;
		ensureMatrixParamsProcessed();
			HDSBuilder b=new HDSBuilder();
			IHDSNodeList nl=mMatrixParam.getNodes("/*");
			for(IHDSNode n: nl)
			{	b.addNode(n.getName(), SecurityUtils.HTMLEncode((String)n.getValue()));					
			}
			response=b.getRoot();
		return response;
	}

	private Object getMethod()
	{
		Object response;
		response=mRequest.getMethod();
		return response;
	}

	private Object getURLRelative() throws MalformedURLException
	{
		Object response;
		URL u=new URL(mURL);
			String rel=u.getPath();
			String query=mRequest.getQueryString();
			if(query!=null)
			{	rel+="?"+query;					
			}
			response=rel;
		return response;
	}

	private Object getURLFull()
	{
		Object response;
		String url=mRequest.getRequestURL().toString();
			String query=mRequest.getQueryString();
			if(query!=null)
			{	url+="?"+query;					
			}
			response=url;
		return response;
	}

	private Object getURLPath() throws MalformedURLException
	{
		Object response;
		URL u=new URL(mURL);
			response=u.getPath();
		return response;
	}

	private Object getURLQuery()
	{
		Object response;
		response=mRequest.getQueryString();
		return response;
	}

	private Object getURL()
	{
		Object response;
		response = mURL;
		return response;
	}

	private Object getParamSpace(String type) throws IOException
	{
		Object response;
		ensureParamsProcessed();
		ensureMatrixParamsProcessed();
		String param = type.substring(HTTP_PARAM_SPACE.length());
		response=mParam.getFirstValue(param);
		if(response==null)
		{	response=mMatrixParam.getFirstValue(param);					
		}
		return response;
	}

	private Object getParamSpaceHTMLEncoded(String type) throws IOException
	{
		Object response;
		ensureParamsProcessed();
		ensureMatrixParamsProcessed();
		String param = type.substring(HTTP_PARAM_SPACE_HTML_ENCODED.length());
		response=mParam.getFirstValue(param);
		if(response==null)
		{	response=mMatrixParam.getFirstValue(param);					
		}
		if(response!=null)
		{	response=SecurityUtils.HTMLEncode((String)response);
		}
		return response;
	}
	
	private Object getParamSet() throws IOException
	{
		Object response;
		ensureParamsProcessed();
		ensureMatrixParamsProcessed();
		HDSBuilder b = new HDSBuilder();
		b.importChildren(mParam);
		b.importChildren(mMatrixParam);
		response = b.getRoot();
		return response;
	}

	private Object getParamSetHTMLEncoded() throws IOException
	{
		Object response;
		IHDSNode p=(IHDSNode)getParamSet();
		HDSBuilder b=new HDSBuilder();
		IHDSNodeList nl=p.getNodes("/*");
		for(IHDSNode n: nl)
		{	b.addNode(n.getName(), SecurityUtils.HTMLEncode((String)n.getValue()));					
		}
		response=b.getRoot();
		return response;
	}
	
	private void innerProcessWSResponseSink(String aType, Object aValue, INKFRequestContext aHelper) throws NKFException, IOException
	{
		if(!mWS.getOutboundConnection().isOpen())
		{	throw new NKFException("WebSocket not Available");			
		}
		if(aType.startsWith("message"))
		{	String m=aHelper.sourcePrimary(String.class);
			mWS.getOutboundConnection().getRemote().sendString(m);
			mWS.log("MSG_WS_MESSAGE_OUTBOUND");
		}
		if(aType.startsWith("disconnect"))
		{	Boolean d=aHelper.sourcePrimary(Boolean.class);
			if(d)
			{	mWS.getOutboundConnection().disconnect();
				mWS.log("MSG_WS_DISCONNECT_OUTBOUND");
			}
		}
	}
		
	
	private void innerProcessResponseSink(String aType, Object aValue, INKFRequestContext aHelper) throws NKFException
	{
		if (aType.startsWith(HTTP_COOKIE_SINK))
		{	Cookie cookie=transreptIfNeeded(aValue,Cookie.class,aHelper);
			mResponseCookies.put(cookie.getName()+cookie.getPath(), cookie);
		}
		else if(aType.startsWith(HTTP_HEADER))
		{	String name = aType.substring(HTTP_HEADER.length());
			//String value=aValue.toString(); //transreptIfNeeded(aValue,String.class,aHelper);
			mResponseHeaders.put(name, aValue);
		}
		else if(aType.startsWith(HTTP_RESPONSE_CODE))
		{	Integer cd=transreptIfNeeded(aValue,Integer.class,aHelper);
			mResponseCode=cd.intValue();
		}
		else if(aType.startsWith(HTTP_REDIRECT))
		{	mResponseRedirect=transreptIfNeeded(aValue,String.class,aHelper);
		}
	}
	
	private <T> T transreptIfNeeded(Object aRep, Class<T> aTo, INKFRequestContext aHelper) throws NKFException
	{	Object result=aRep;
		if (!aTo.isAssignableFrom(aRep.getClass()))
		{	result=aHelper.transrept(aRep, aTo);
		}
		return (T)result;
	}
	
	
	private void ensureParamsProcessed() throws IOException
	{	if (mParam==null)
		{	mParam=sFailedParam;
		
			String encoding=mRequest.getCharacterEncoding();
			String mime=mRequest.getContentType();
			int hash=0;
			
			HDSBuilder b = new HDSBuilder();
			
			if (mime!=null && mime.startsWith("multipart/form-data"))
			{	// multipart content = uploads
				MultiPartRequest mpr=new MultiPartRequest(mRequest);
				String[] names=mpr.getPartNames();
				String localEncoding=encoding;
				if(localEncoding==null)
				{	localEncoding=mConfig.getDefaultPostParamEncoding();
				}
				for(int i=0; i<names.length;i++)
				{	String name=names[i];
					String filename=mpr.getFilename(name);
					if(filename!=null)
					{	b.addNode(name,filename);
						if (mUploads==null)
						{	mUploads = new HashMap();
						}
						mUploads.put(name,new ParameterUploadAspect(mpr,name));
					}
					else
					{	// must be form data in addition to the upload
						InputStream is=mpr.getInputStream(name);
						InputStreamReader isr=new InputStreamReader(is, localEncoding); 	//Maybe should try to get the encoding from the MPR - but it has no access - so we use the mRequest's value above
						StringWriter sw=new StringWriter(mpr.getContentLength(name));
						while(isr.ready())
						{	sw.write(isr.read());
						}
						String value=sw.toString();
						hash^=value.hashCode();
						b.addNode(name,value);
					}
				}
			}
			
			if (mime==null || (mime!=null && !mime.startsWith("application/x-www-form-urlencoded") && !mime.startsWith("multipart/form-data")))
			{	// special body content
				
				if (mRequest.getContentLength()>0)
				{	ByteArrayOutputStream baos = new ByteArrayOutputStream(mRequest.getContentLength());
					Utils.pipe(mRequest.getInputStream(), baos);
					mContent = new ByteArrayRepresentation(baos, mRequest.getCharacterEncoding());
				}
			}
			
			// fix to work around unknown encoding of post parameters
			// i.e. the standard says they should be same as the encoding of the HTML form that contains the FORM
			// - not very useful
			final String JETTY_ASSUMED_ENCODING="UTF-8";
			String assumedEncoding=mConfig.getDefaultPostParamEncoding();
			boolean unmanglePostParams= (mRequest.getMethod().equals("POST") && encoding==null  &&  !assumedEncoding.equals(JETTY_ASSUMED_ENCODING));
			
			Enumeration<String> i=mRequest.getParameterNames();
			ArrayList<String> paramNames=new ArrayList<String>(16);
			while(i.hasMoreElements())
			{	paramNames.add(i.nextElement());
			}
			Collections.sort(paramNames);
			for (String name : paramNames)
			{	for (String value: mRequest.getParameterValues(name) )
				{	if (unmanglePostParams)
					{	byte[] b2=value.getBytes(JETTY_ASSUMED_ENCODING);
						value=new String(b2,assumedEncoding);
					}
					b.addNode(name, value);
					hash^=value.hashCode();

				}
			}
			
			mParam = b.getRoot();
			mParamHash=hash;
		}
	}
	
	private void ensureMatrixParamsProcessed() throws IOException
	{	if (mMatrixParam==null)
		{	HDSBuilder b=new HDSBuilder();
			String[] tokens1 = mURL.split(";");
			for(int i=0; i<tokens1.length; i++)
			{	String[] tokens2 = tokens1[i].split("=");
				if(tokens2.length==2)
				{	b.addNode(tokens2[0], tokens2[1]);
				}
			}
			mMatrixParam=b.getRoot();
		}
	}
	
	public boolean resolveEquals(Object aOther)
	{	return aOther instanceof HTTPRequestSpace2Wrapper;
	}
	
	public Map getResponseCookies()
	{	return mResponseCookies;
	}
	public Map getResponseHeaders()
	{	return mResponseHeaders;
	}
	public int getResponseCode()
	{	return mResponseCode;		
	}
	public String getResponseRedirect()
	{	return mResponseRedirect;		
	}
	public boolean shouldRedirect()
	{	return mResponseRedirect!=null && mResponseCode==200;
	}
	
	void applyResponseMetadata(IRequestResponseFields aResponseMetadata, INKFRequestContext aContext) throws NKFException
	{	for (int i=0; i<aResponseMetadata.size(); i++)
		{	String key=aResponseMetadata.getKey(i);
			if (key.startsWith(HTTP_RESPONSE_SCHEME))
			{	key=key.substring(HTTP_RESPONSE_SCHEME.length());
				Object value=aResponseMetadata.getValue(i);
				innerProcessResponseSink(key,value,aContext);
			}
		}
	}
	
	
	/* equals optimization */
	private boolean mBasicAccess;
	private boolean mAdvancedAccess;
	
	public boolean equals(Object aOther)
	{	boolean result;
		if (aOther instanceof HTTPRequestSpace2Wrapper)
		{	HTTPRequestSpace2Wrapper other=(HTTPRequestSpace2Wrapper)aOther;
			if (aOther==this)
			{	result=true;
			}
			else if (mAdvancedAccess || other.mAdvancedAccess)
			{	result=false;
			}
			else if (!mBasicAccess && !other.mBasicAccess)
			{	result=true;
			}
			else
			{	result=innerEquals(other);
			}
		}
		else
		{	result=false;
		}
		return result;
	}
	
	private boolean innerEquals(HTTPRequestSpace2Wrapper aOther)
	{	boolean result=mURL.equals(aOther.mURL);
		if (result)
		{	result=mMethod.equals(aOther.mMethod);
			if (result)
			{	try
				{	mBasicAccess=true;
					aOther.mBasicAccess=true;
					ensureParamsProcessed();
					aOther.ensureParamsProcessed();
					ensureMatrixParamsProcessed();
					aOther.ensureMatrixParamsProcessed();
				} catch (Exception e) {}
				if (mParamHash==aOther.mParamHash)
				{	try
					{	result=getCollection().equals(aOther.getCollection());		//Use HDS comparison on the collection of all readable parameters
					}
					catch(Exception e)
					{	result=false;						
					}
					/*
					IHDSNode[] ch1=mParam.getChildren();
					IHDSNode[] ch2=aOther.mParam.getChildren();
					if (ch1.length==ch2.length)
					{	for (int i=0; i<ch1.length; i++)
						{	if (!ch1[i].getName().equals(ch2[i].getName()) || !ch1[i].getValue().equals(ch2[i].getValue()))
							{	result=false;
								break;
							}
						}
					}
					else
					{	result=false;
					}
					*/
					
				}
				else
				{	result=false;
				}
			}
		}
		return result;
	}
	
	public int hashCode()
	{	return mHash;
	}
	
}