package org.netkernel.http.transport;

import java.lang.reflect.Method;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

public abstract class HttpAccessorImpl extends StandardAccessorImpl {
	private enum HttpMethod
	{	GET("onGet"),
		POST("onPost"),
		DELETE("onDelete"),
		PUT("onPut"),
		PATCH("onPatch"),
		HEAD("onHead");

		String methodName;
		private HttpMethod(String methodName)
		{	this.methodName = methodName;
		}
  };
  
  @Override
  public void onSource(INKFRequestContext aContext) throws Exception
  {	String method = aContext.source("httpRequest:/method", String.class);
  	switch (HttpMethod.valueOf(method))
    {
		case GET:
			onGet(aContext);
			break;
		case POST:
			onPost(aContext);
			break;
		case DELETE:
			onDelete(aContext);
			break;
		case PUT:
			onPut(aContext);
			break;
		case PATCH:
			onPatch(aContext);
			break;
		case HEAD:
			onHead(aContext);
			break;
		default:
			throw new Exception("501 Unknown HTTP method: " + method);
    }
  }

  private void unsupportedMethod(INKFRequestContext aContext) throws NKFException
  {	String supportedMethods = "";
	String nextSep= "";
	for (int i= 0; i < HttpMethod.values().length; i++)
	{	if (methodSupported(HttpMethod.values()[i].methodName))
		{	supportedMethods += nextSep + HttpMethod.values()[i].name();
			nextSep= ", ";
		}
	}
	aContext.sink("httpResponse:/code", 405);
	aContext.sink("httpResponse:/header/Allow", supportedMethods);
	aContext.createResponseFrom("405: Unsupported Method");
  }
  
  private boolean methodSupported(String methodName)
  {
	try
	{	Method method= this.getClass().getMethod(methodName, INKFRequestContext.class);
		return !method.getDeclaringClass().equals(HttpAccessorImpl.class);
	}
	catch (Exception e)
	{	return false;
	}
  }
  
  public void onGet(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
  public void onPost(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
  public void onDelete(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
  public void onPut(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
  public void onPatch(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
  public void onHead(INKFRequestContext aContext) throws Exception
  {	unsupportedMethod(aContext);
  }
}