package org.netkernel.http.transport.representation;

import java.net.URI;
import java.util.UUID;

import org.eclipse.jetty.websocket.api.Session;
import org.eclipse.jetty.websocket.api.annotations.OnWebSocketClose;
import org.eclipse.jetty.websocket.api.annotations.OnWebSocketConnect;
import org.eclipse.jetty.websocket.api.annotations.OnWebSocketMessage;
import org.eclipse.jetty.websocket.api.annotations.WebSocket;
import org.netkernel.http.transport.HTTPRequestSpace2Wrapper;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.util.RegexRewriterRule;

/********
 * This interface now needs to be done with Annotations...
 * http://www.eclipse.org/jetty/documentation/9.2.6.v20141205/jetty-websocket-api-annotations.html
 * @author pjr
 *
 */
@WebSocket
public class NKWebSocket
{
	private HTTPRequestSpace2Wrapper mTransportSpace;
	private String mResourceIdentifier;
	
	private Session mConnection;
	
	private String mRemoteAddr;
	private String mSocketID;
	
	private String mProtocol;
	
	public NKWebSocket(String aURL, String aRemoteAddr, String aProtocol)
	{	
		mResourceIdentifier=generateWSURLFromURL(aURL);
		mRemoteAddr=aRemoteAddr;
		mProtocol=aProtocol;
	}
	
	public static String generateWSURLFromURL(String aURL)
	{	URI u=URI.create(aURL);
		return "ws://"+u.getHost()+u.getPath();
	}
	
	public String getProtocol()
	{	return mProtocol;		
	}
	
	public void setTransportSpace(HTTPRequestSpace2Wrapper aTransportSpace)
	{	mTransportSpace=aTransportSpace;
		RegexRewriterRule[] rules=mTransportSpace.getRewrtieRules();
		if(rules!=null && rules.length>0)
		{	for(RegexRewriterRule rule : rules )
			{	if(rule.matches(mResourceIdentifier))
				{	mResourceIdentifier=rule.map(mResourceIdentifier);
				}
			}
		}
	}
	
	@OnWebSocketClose
	public void onClose(int arg0, String arg1)
	{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_WS_DISCONNECT", mRemoteAddr, mResourceIdentifier);
		try
		{	if(mSocketID!=null)
			{	INKFRequestContext context=mTransportSpace.getTransportContext();
				INKFRequest req=context.createRequest(mResourceIdentifier);
				req.addArgument("socketid", mSocketID);
				req.setVerb(INKFRequestReadOnly.VERB_DELETE);
				context.issueRequest(req);
			}
		}
		catch(Exception e)
		{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING, "EX_WS_DISCONNECT", mRemoteAddr, mResourceIdentifier, e);
		}
		mTransportSpace.disconnectWS();
	}

	@OnWebSocketConnect
	public void onOpen(Session aConnection)
	{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_WS_CONNECT", mRemoteAddr, mResourceIdentifier);
		mConnection=aConnection;
		try
		{
			INKFRequestContext context=mTransportSpace.getTransportContext();
			INKFRequest req=context.createRequest(mResourceIdentifier);
			req.setVerb(INKFRequestReadOnly.VERB_NEW);
			req.addPrimaryArgument(mConnection);
			req.setRepresentationClass(String.class);
			String id=(String)context.issueRequest(req);
			if(id==null)
			{	id=UUID.randomUUID().toString();				
			}
			mSocketID=id;
		}
		catch(Exception e)
		{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING, "EX_WS_CONNECT", mRemoteAddr, mResourceIdentifier, e);		
		}
	}

	public Session getOutboundConnection()
	{	return mConnection;		
	}
	
	public void log(String id)
	{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_FINE, id, mRemoteAddr, mResourceIdentifier);
	}

	@OnWebSocketMessage
	public void onMessage(byte[] arg0, int arg1, int arg2)
	{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING, "MSG_WS_MESSAGE", mRemoteAddr, "onMessage(byte,byte,int,int) Not Supported");
	}

	@OnWebSocketMessage
	public void onMessage(String message)
	{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_FINE, "MSG_WS_MESSAGE", mRemoteAddr, mResourceIdentifier);
		try
		{	INKFRequestContext context=mTransportSpace.getTransportContext();
			INKFRequest req=context.createRequest(mResourceIdentifier);
			req.addArgument("socketid", mSocketID);
			req.setVerb(INKFRequestReadOnly.VERB_SINK);
			req.addPrimaryArgument(message);
			context.issueRequest(req);
		}
		catch(Exception e)
		{	mTransportSpace.getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING, "EX_WS_MESSAGE", mRemoteAddr, mResourceIdentifier, e);
		}
	}
}
