package org.netkernel.http.util;

import java.util.*;

import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.http.HttpHeader;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;

public class AcceptHeaderParser
{
	public static IHDSNode getAcceptHeaderState(HttpServletRequest aRequest)
	{	
		HDSBuilder b=new HDSBuilder();
		b.pushNode("headers");
		processHeader(b,HttpHeader.ACCEPT.asString(),aRequest.getHeader(HttpHeader.ACCEPT.asString()),true);
		processHeader(b,HttpHeader.ACCEPT_CHARSET.asString(),aRequest.getHeader(HttpHeader.ACCEPT_CHARSET.asString()),false);
		processHeader(b,HttpHeader.ACCEPT_ENCODING.asString(),aRequest.getHeader(HttpHeader.ACCEPT_ENCODING.asString()),false);
		processHeader(b,HttpHeader.ACCEPT_LANGUAGE.asString(),aRequest.getHeader(HttpHeader.ACCEPT_LANGUAGE.asString()),false);
		return b.getRoot();
	}
	
	private static void processHeader(HDSBuilder b, String aTag, String aHeaderValue, boolean aMediaRangeSort)
	{
		if (aHeaderValue==null || aHeaderValue.length()==0) return;
		String h=aHeaderValue;
		List<AcceptEntry> entries=new ArrayList<AcceptEntry>();
		int i=0;
		int l=h.length();
		int n=0;
		while (i<l)
		{	int p2=-1;
			for (int p=i; i<l; p++)
			{	char c=p<l?h.charAt(p):' ';
				if (c==',' || p==l)
				{	float q=1.0f;
					String value;
					if (p2>0)
					{	String qString=h.substring(p2+1,p).trim(); 
						if (qString.startsWith("q="))
						{	try
							{	q=Float.parseFloat(qString.substring(2));
								if (q<0.0f) q=0.0f;
								else if (q>1.0f) q=1.0f;
							} catch(NumberFormatException e) {;}
							value=h.substring(i,p2);
						}
						else
						{	value=h.substring(i,p);
						}
					}
					else
					{	value=h.substring(i,p);
					}
					value=value.trim();
					int mediaRangeDepth=aMediaRangeSort?getMediaRangeDepth(value):0;
					entries.add(new AcceptEntry(value, q, n++,mediaRangeDepth));
					i=p+1;
					break;
				}
				else if (c==';')
				{	p2=p;
				}
			}
		}
		
		if (entries.size()>0)
		{	Collections.sort(entries);
			b.pushNode(aTag);
			for (AcceptEntry ae: entries)
			{	b.pushNode("entry");
				b.addNode("value",ae.getValue());
				b.addNode("q", ae.getScore());
				b.popNode();
			}
			b.popNode();
		}
	}
	
	private static int getMediaRangeDepth(String aMediaRange)
	{
		int result;
		if (aMediaRange.startsWith("*"))
		{	result=0;
		}
		else
		{	int i=aMediaRange.indexOf('/');
			String subRange=aMediaRange.substring(i+1);
			if (subRange.startsWith("*"))
			{	result=1;
			}
			else
			{	i=subRange.indexOf(';');
				if (i>=0 && i<subRange.length()-1)
				{	result=3;
				}
				else
				{	result=2;
				}
			}
		}
		return result;
	}
	
	
	private static class AcceptEntry implements Comparable<AcceptEntry>
	{
		private final String mValue;
		private final Float mScore;
		private final Integer mOrder;
		private final Integer mMediaRangeDepth;
		
		public AcceptEntry(String aValue, float aScore, int aOrder, int aMediaRangeDepth)
		{	mValue=aValue;
			mScore=aScore;
			mOrder=aOrder;
			mMediaRangeDepth=aMediaRangeDepth;
		}
		
		public String toString()
		{	return mValue+"="+mScore;
		}
		
		public String getValue()
		{	return mValue;
		}
		
		public Float getScore()
		{	return mScore;
		}

		public int compareTo(AcceptEntry o)
		{
			//sort order
			// 1 score
			// 2 depth of mime
			// 3 order  (first is highest)
			if (!o.mScore.equals(mScore)) return o.mScore.compareTo(mScore);
			if (!o.mMediaRangeDepth.equals(mMediaRangeDepth)) return o.mMediaRangeDepth.compareTo(mMediaRangeDepth);
			return o.mOrder<mOrder? o.mOrder : mOrder;
		}
	}
}